@@ -1,20 +1,20 @@
 import os
 import re
-import subprocess
 import sys
 import time
 import shutil
+import contextlib
 
-from typing import List, Iterable, Dict, Tuple, Callable, Any, Optional
+from typing import List, Iterable, Dict, Tuple, Callable, Any, Optional, Iterator
 
 from mypy import defaults
-from mypy.test.config import test_temp_dir
+import mypy.api as api
 
 import pytest  # type: ignore  # no pytest in typeshed
 
 # Exporting Suite as alias to TestCase for backwards compatibility
 # TODO: avoid aliasing - import and subclass TestCase directly
-from unittest import TestCase as Suite
+from unittest import TestCase as Suite  # noqa: F401 (re-exporting)
 
 from mypy.main import process_options
 from mypy.options import Options
@@ -27,14 +27,30 @@ skip = pytest.mark.skip
 MIN_LINE_LENGTH_FOR_ALIGNMENT = 5
 
 
+def run_mypy(args: List[str]) -> None:
+    __tracebackhide__ = True
+    outval, errval, status = api.run(args + ['--show-traceback',
+                                             '--no-site-packages',
+                                             '--no-silence-site-packages'])
+    if status != 0:
+        sys.stdout.write(outval)
+        sys.stderr.write(errval)
+        pytest.fail(msg="Sample check failed", pytrace=False)
+
+
 def assert_string_arrays_equal(expected: List[str], actual: List[str],
                                msg: str) -> None:
     """Assert that two string arrays are equal.
 
+    We consider "can't" and "cannot" equivalent, by replacing the
+    former with the latter before comparing.
+
     Display any differences in a human-readable form.
     """
 
     actual = clean_up(actual)
+    actual = [line.replace("can't", "cannot") for line in actual]
+    expected = [line.replace("can't", "cannot") for line in expected]
 
     if actual != expected:
         num_skip_start = num_skipped_prefix_lines(expected, actual)
@@ -81,14 +97,14 @@ def assert_string_arrays_equal(expected: List[str], actual: List[str],
                 if len(a) > width:
                     sys.stderr.write('...')
             sys.stderr.write('\n')
-        if actual == []:
+        if not actual:
             sys.stderr.write('  (empty)\n')
         if num_skip_end > 0:
             sys.stderr.write('  ...\n')
 
         sys.stderr.write('\n')
 
-        if first_diff >= 0 and first_diff < len(actual) and (
+        if 0 <= first_diff < len(actual) and (
                 len(expected[first_diff]) >= MIN_LINE_LENGTH_FOR_ALIGNMENT
                 or len(actual[first_diff]) >= MIN_LINE_LENGTH_FOR_ALIGNMENT):
             # Display message that helps visualize the differences between two
@@ -113,10 +129,24 @@ def assert_module_equivalence(name: str,
                  name))
 
 
+def assert_target_equivalence(name: str,
+                              expected: Optional[List[str]], actual: List[str]) -> None:
+    """Compare actual and expected targets (order sensitive)."""
+    if expected is not None:
+        assert_string_arrays_equal(
+            expected,
+            actual,
+            ('Actual targets ({}) do not match expected targets ({}) '
+             'for "[{} ...]"').format(
+                 ', '.join(actual),
+                 ', '.join(expected),
+                 name))
+
+
 def update_testcase_output(testcase: DataDrivenTestCase, output: List[str]) -> None:
     assert testcase.old_cwd is not None, "test was not properly set up"
     testcase_path = os.path.join(testcase.old_cwd, testcase.file)
-    with open(testcase_path) as f:
+    with open(testcase_path, encoding='utf8') as f:
         data_lines = f.read().splitlines()
     test = '\n'.join(data_lines[testcase.line:testcase.lastline])
 
@@ -140,7 +170,7 @@ def update_testcase_output(testcase: DataDrivenTestCase, output: List[str]) -> N
 
     data_lines[testcase.line:testcase.lastline] = [test]
     data = '\n'.join(data_lines)
-    with open(testcase_path, 'w') as f:
+    with open(testcase_path, 'w', encoding='utf8') as f:
         print(data, file=f)
 
 
@@ -215,6 +245,22 @@ def clean_up(a: List[str]) -> List[str]:
     return res
 
 
+@contextlib.contextmanager
+def local_sys_path_set() -> Iterator[None]:
+    """Temporary insert current directory into sys.path.
+
+    This can be used by test cases that do runtime imports, for example
+    by the stubgen tests.
+    """
+    old_sys_path = sys.path[:]
+    if not ('' in sys.path or '.' in sys.path):
+        sys.path.insert(0, '')
+    try:
+        yield
+    finally:
+        sys.path = old_sys_path
+
+
 def num_skipped_prefix_lines(a1: List[str], a2: List[str]) -> int:
     num_eq = 0
     while num_eq < min(len(a1), len(a2)) and a1[num_eq] == a2[num_eq]:
@@ -339,12 +385,21 @@ def parse_options(program_text: str, testcase: DataDrivenTestCase,
             raise RuntimeError('Specifying targets via the flags pragma is not supported.')
     else:
         options = Options()
+        # TODO: Enable strict optional in test cases by default (requires *many* test case changes)
+        options.strict_optional = False
 
     # Allow custom python version to override testcase_pyversion
     if (not flag_list or
             all(flag not in flag_list for flag in ['--python-version', '-2', '--py2'])):
         options.python_version = testcase_pyversion(testcase.file, testcase.name)
 
+    if testcase.config.getoption('--mypy-verbose'):
+        options.verbosity = testcase.config.getoption('--mypy-verbose')
+
+    if os.getenv('NEWSEMANAL'):
+        if not flag_list or '--no-new-semantic-analyzer' not in flag_list:
+            options.new_semantic_analyzer = True
+
     return options
 
 
@@ -357,24 +412,6 @@ def split_lines(*streams: bytes) -> List[str]:
     ]
 
 
-def run_command(cmdline: List[str], *, env: Optional[Dict[str, str]] = None,
-                timeout: int = 300, cwd: str = test_temp_dir) -> Tuple[int, List[str]]:
-    """A poor man's subprocess.run() for 3.4 compatibility."""
-    process = subprocess.Popen(
-        cmdline,
-        env=env,
-        stdout=subprocess.PIPE,
-        stderr=subprocess.PIPE,
-        cwd=cwd,
-    )
-    try:
-        out, err = process.communicate(timeout=timeout)
-    except subprocess.TimeoutExpired:
-        out = err = b''
-        process.kill()
-    return process.returncode, split_lines(out, err)
-
-
 def copy_and_fudge_mtime(source_path: str, target_path: str) -> None:
     # In some systems, mtime has a resolution of 1 second which can
     # cause annoying-to-debug issues when a file has the same size
