@@ -1,4 +1,4 @@
-import glob
+import io
 import os.path
 import shutil
 import sys
@@ -6,15 +6,18 @@ import tempfile
 import re
 from types import ModuleType
 
-from typing import List, Tuple
+from typing import Any, List, Tuple, Optional
 
-from mypy.test.helpers import Suite, assert_equal, assert_string_arrays_equal
+from mypy.test.helpers import (
+    Suite, assert_equal, assert_string_arrays_equal, local_sys_path_set
+)
 from mypy.test.data import DataSuite, DataDrivenTestCase
 from mypy.errors import CompileError
 from mypy.stubgen import (
-    generate_stubs, parse_options, walk_packages, Options, collect_build_targets,
+    generate_stubs, parse_options, Options, collect_build_targets,
     mypy_options
 )
+from mypy.stubutil import walk_packages
 from mypy.stubgenc import generate_c_type_stub, infer_method_sig, generate_c_function_stub
 from mypy.stubdoc import (
     parse_signature, parse_all_signatures, build_signature, find_unique_signatures,
@@ -63,11 +66,30 @@ class StubgenCmdLineSuite(Suite):
             finally:
                 os.chdir(current)
 
+    def test_module_not_found(self) -> None:
+        current = os.getcwd()
+        captured_output = io.StringIO()
+        sys.stdout = captured_output
+        with tempfile.TemporaryDirectory() as tmp:
+            try:
+                os.chdir(tmp)
+                self.make_file(tmp, 'mymodule.py', content='import a')
+                opts = parse_options(['-m', 'mymodule'])
+                py_mods, c_mods = collect_build_targets(opts, mypy_options(opts))
+                self.assertRegex(captured_output.getvalue(), "No module named 'a'")
+            finally:
+                sys.stdout = sys.__stdout__
+                os.chdir(current)
+
     def make_file(self, *path: str, content: str = '') -> None:
         file = os.path.join(*path)
         with open(file, 'w') as f:
             f.write(content)
 
+    def run(self, result: Optional[Any] = None) -> Optional[Any]:
+        with local_sys_path_set():
+            return super().run(result)
+
 
 class StubgenCliParseSuite(Suite):
     def test_walk_packages(self) -> None:
@@ -218,6 +240,19 @@ class StubgenUtilSuite(Suite):
 
         assert_equal(infer_sig_from_docstring('\nfunc[(x: foo.bar, invalid]', 'func'), [])
 
+        assert_equal(infer_sig_from_docstring('\nfunc(x: invalid::type<with_template>)', 'func'),
+                     [FunctionSig(name='func', args=[ArgSig(name='x', type=None)],
+                                  ret_type='Any')])
+
+        assert_equal(infer_sig_from_docstring('\nfunc(x: str="")', 'func'),
+                     [FunctionSig(name='func', args=[ArgSig(name='x', type='str', default=True)],
+                                  ret_type='Any')])
+
+    def test_infer_sig_from_docstring_duplicate_args(self) -> None:
+        assert_equal(infer_sig_from_docstring('\nfunc(x, x) -> str\nfunc(x, y) -> int', 'func'),
+                     [FunctionSig(name='func', args=[ArgSig(name='x'), ArgSig(name='y')],
+                                  ret_type='int')])
+
     def test_infer_arg_sig_from_docstring(self) -> None:
         assert_equal(infer_arg_sig_from_docstring("(*args, **kwargs)"),
                      [ArgSig(name='*args'), ArgSig(name='**kwargs')])
@@ -242,6 +277,10 @@ class StubgenPythonSuite(DataSuite):
     files = ['stubgen.test']
 
     def run_case(self, testcase: DataDrivenTestCase) -> None:
+        with local_sys_path_set():
+            self.run_case_inner(testcase)
+
+    def run_case_inner(self, testcase: DataDrivenTestCase) -> None:
         extra = []
         mods = []
         source = '\n'.join(testcase.input)
@@ -380,6 +419,21 @@ class StubgencSuite(Suite):
         assert_equal(output, ['def test(self, arg0: int) -> Any: ...'])
         assert_equal(imports, [])
 
+    def test_generate_c_type_with_docstring_empty_default(self) -> None:
+        class TestClass:
+            def test(self, arg0: str = "") -> None:
+                """
+                test(self: TestClass, arg0: str = "")
+                """
+                pass
+        output = []  # type: List[str]
+        imports = []  # type: List[str]
+        mod = ModuleType(TestClass.__module__, '')
+        generate_c_function_stub(mod, 'test', TestClass.test, output, imports,
+                                 self_var='self', class_name='TestClass')
+        assert_equal(output, ['def test(self, arg0: str = ...) -> Any: ...'])
+        assert_equal(imports, [])
+
     def test_generate_c_function_other_module_arg(self) -> None:
         """Test that if argument references type from other module, module will be imported."""
         # Provide different type in python spec than in docstring to make sure, that docstring
